/**
 * Netwosix Web Admin - libal
 * Copyright (C) 2006 netwosix team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * If you want to contact e-mail at info@netwosix.org
 */


#include "al.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>


/**
 */
struct _AlSocketManager
{
	int                init;

	int                sock;
	int                data;
	int                on;
	struct sockaddr_in name;
	unsigned long      ioctlarg;
	int                port;

};

typedef struct _AlSocketManager AlSocketManager;


/**
 */
struct _AlSocket
{

	int                  sock;
	struct sockaddr_in   sn;
	socklen_t            snsize;
	struct hostent     * hostname;

};

typedef struct _AlSocket AlSocket;


/**
 */
void al_socket_connection_manager_delete(AlConnectionManager * connection_manager)
{

	al_return_if_fail (connection_manager);

	al_delete(connection_manager->data);
	al_delete(connection_manager);

}


/**
 */
AlConnectionManager * al_socket_connection_manager_new(int port)
{
	AlConnectionManager * connection = al_new(AlConnectionManager);
	AlSocketManager     * socket_manager = al_new(AlSocketManager);
 

	connection->wait = al_socket_connection_wait;
	connection->close = al_socket_connection_close;
	connection->read = al_socket_connection_read;
	connection->send = al_socket_connection_send;
	connection->del = al_socket_connection_manager_delete;
	connection->conn = al_socket_connection_connect;

	connection->data = socket_manager;
	connection->fork = 1;

	socket_manager->init = 0;
	socket_manager->data = 0;
 	socket_manager->on = 1;
	socket_manager->ioctlarg = 0;
	socket_manager->port = port;


	return connection;
}


/**
 */
AlConnection * al_socket_connection_wait(AlConnectionManager * connection_manager)
{
	AlConnection     * ret = NULL;
	AlSocket         * ret_sock = NULL;
	AlSocketManager  * socket_manager;



	al_return_val_if_fail(connection_manager, NULL);

	socket_manager = connection_manager->data;

	if (!socket_manager->init)
	{

		socket_manager->sock = socket(PF_INET, SOCK_STREAM, 0);
		if (socket_manager->sock == -1)
		{
			syslog(LOG_ERR, "Could not create the socket.\n");
			return NULL;
		}

		/* Give the socket a name. */
		socket_manager->name.sin_family = AF_INET;
		socket_manager->name.sin_port = htons (socket_manager->port);
		socket_manager->name.sin_addr.s_addr = htonl (INADDR_ANY);
		setsockopt(socket_manager->sock, SOL_SOCKET, SO_REUSEADDR, &(socket_manager->on), sizeof(socket_manager->on));
		//ioctl(sock, FIONBIO, &ioctlarg);
		if (bind (socket_manager->sock, (struct sockaddr *) &(socket_manager->name), sizeof (socket_manager->name)) < 0)
		{
			syslog (LOG_ERR, "Could not bind the server. Error %d\n", errno);
			return NULL;
		}

		if (listen (socket_manager->sock, 1) < 0)
		{
			syslog (LOG_ERR, "Could not start the listener.\n");
			return NULL;
		}



		syslog(LOG_NOTICE, "Acceptiong connections on port %d\n", socket_manager->port);

		socket_manager->init = 1;
	}



	ret = al_new(AlConnection);
	ret->connection_manager = connection_manager;

	ret_sock = al_new(AlSocket);
	ret_sock->snsize = sizeof(ret_sock->sn);
	ret->data = ret_sock;


	ret_sock->sock = accept(socket_manager->sock, (struct sockaddr *)&(ret_sock->sn), &(ret_sock->snsize));

	if(ret_sock->sock == -1)
	{
		al_delete(ret);
		al_delete(ret_sock);

		syslog(LOG_NOTICE, "Close connections.\n");

		return NULL;
	}

	ret_sock->hostname = gethostbyaddr(&(ret_sock->sn.sin_addr), sizeof(ret_sock->sn.sin_addr), AF_INET);
	if (ret_sock->hostname) syslog(LOG_NOTICE, "Accepted new connection from %s.\n", ret_sock->hostname->h_name);


	return ret;
}


/**
 */
void al_socket_connection_close(AlConnection * connection)
{
	al_return_if_fail(connection);

	al_delete(connection->data);
	al_delete(connection);
}


/**
 */
AlString * al_socket_connection_read(AlConnection * connection)
{
	AlString    * ret = NULL;
	AlString    * s = NULL;
	AlSocket    * socket;
	const int     len = 3;
	int           i;
	char          buf[len];
	int           rec;


	al_return_val_if_fail(connection, NULL);

	socket = connection->data;
	s = al_string_new();
	
	while ((rec = recv(socket->sock, buf, sizeof(char) * len, 0)) > 0)
	{

		if (rec == 3 && buf[0] == 'E' && buf[1] == 'O' && buf[2] == 'P')
		{
			// End of request packet
			ret = s;
			break;
		}
		else if (rec == 3 && buf[0] == 'E' && buf[1] == 'O' && buf[2] == 'F')
		{
			// End of trasmisson
			ret = NULL;
			if (socket->hostname) syslog(LOG_NOTICE, "Close connection from %s.\n", socket->hostname->h_name);
			shutdown(socket->sock, SHUT_RDWR);
			break;
		}


		for (i=0;i < rec;i++) al_string_append_char(s, buf[i]);

	}


	if (ret == NULL && s)
	{
		// Incomplete data. discard.
		al_string_delete(s);
	}


	return ret;
}


/**
 */
void al_socket_connection_send(AlConnection * connection, const AlString * response)
{
	AlSocket    * socket;


	al_return_if_fail(connection && response);

	socket = connection->data;

	if (send(socket->sock, al_string_get(response), al_string_len(response), 0) == -1)
	{
		syslog(LOG_ERR, "Could not send data %s\n", al_string_get(response));
	}

	if (send(socket->sock, "EOP", strlen("EOP"), 0) == -1)
	{
		syslog(LOG_ERR, "Could not send data %s\n", "EOP");
	}

}


/**
 */
AlConnection * al_socket_connection_connect(AlConnectionManager * connection_manager, const char * server)
{
	AlSocketManager      * socket_manager;
	AlConnection         * ret;
	AlSocket             * ret_sock;
	struct hostent       * pent;



	al_return_val_if_fail(connection_manager && server, NULL);
	
	socket_manager = connection_manager->data;

	ret = al_new(AlConnection);
	ret->connection_manager = connection_manager;
	ret_sock = al_new(AlSocket);
	ret->data = ret_sock;

	ret_sock->snsize = sizeof(ret_sock->sn);

	ret_sock->sn.sin_family = AF_INET;

	/* Figure out the internet address. */
	ret_sock->sn.sin_addr.s_addr = inet_addr(server);
	
	if (ret_sock->sn.sin_addr.s_addr == INADDR_NONE)
	{
		pent = gethostbyname(server);
		if (pent)
		{
			ret_sock->sn.sin_addr.s_addr = *((unsigned long *) (pent->h_addr));
		}
		else
		{
			al_log (("Could not resolve the host %s\n", server));
			al_delete(ret_sock);
			al_delete(ret);
			return NULL;
		}
	}

	ret_sock->sock = socket(PF_INET, SOCK_STREAM, 0);
	if (ret_sock->sock == -1)
	{
		al_log(("Could not create the socket.\n"));
		al_delete(ret_sock);
		al_delete(ret);
		return NULL;
	}

	/* Figure out the port name. */
	ret_sock->sn.sin_port =  htons(socket_manager->port);

	if (connect(ret_sock->sock, (struct sockaddr*)&ret_sock->sn, ret_sock->snsize) == -1)
	{
		al_log (("Could not connect to %s at port %d\n", server, socket_manager->port));
		al_delete(ret_sock);
		al_delete(ret);
		ret = NULL;
	}


	return ret;
}
