/**
 * Netwosix Web Admin - libal
 * Copyright (C) 2006 netwosix team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * If you want to contact e-mail at info@netwosix.org
 */


#include "al.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <fcntl.h>
#include <signal.h>
#include <openssl/ssl.h>
#include <openssl/err.h>


#ifndef ALLOW_OLD_VERSIONS
#if (OPENSSL_VERSION_NUMBER < 0x00905100L)
#error "Must use OpenSSL 0.9.6 or later"
#endif
#endif


/**
 */
struct _AlSslManager
{
	int                 init;

	int                 sock;
	int                 data;
	int                 on;
	struct sockaddr_in  name;
	unsigned long       ioctlarg;
	int                 port;

	int                 s_server_session_id_context;
	int                 s_server_auth_session_id_context;

	SSL_CTX           * ctx;

	char              * keyfile;
	char              * dh_file;
	char              * ca_list;
	char              * password;

};

typedef struct _AlSslManager AlSslManager;


/**
 */
struct _AlSsl
{
	
	int                  sock;
	struct sockaddr_in   sn;
	socklen_t            snsize;
	struct hostent     * hostname;
	SSL                * ssl;
	BIO                * sbio;

};

typedef struct _AlSsl AlSsl;


/**
 */
static char * al_ssl_pass;


/**
 */
static int al_ssl_password_cb(char *buf,int num, int rwflag,void *userdata);


/**
 */
static void al_ssl_sigpipe_handle(int x);


/**
 */
SSL_CTX * al_ssl_initialize_ctx(const char * keyfile, const char * ca_list, const char * password);


/**
 */
void al_ssl_destroy_ctx(SSL_CTX *ctx);


/**
 */
void al_ssl_check_cert(SSL * ssl, const char * host);


/**
 */
void al_ssl_load_dh_params(SSL_CTX *ctx, const char * file);


/**
 */
void al_ssl_connection_manager_delete(AlConnectionManager * connection_manager)
{
	AlSslManager        * ssl_manager;



	al_return_if_fail (connection_manager);

	ssl_manager = connection_manager->data;
	if (ssl_manager)
	{
		if (ssl_manager->ctx) al_ssl_destroy_ctx(ssl_manager->ctx);

		al_delete(ssl_manager->keyfile);
		al_delete(ssl_manager->dh_file);
		al_delete(ssl_manager->ca_list);
		al_delete(ssl_manager->password);
	}

	al_delete(connection_manager->data);
	al_delete(connection_manager);

}


/**
 */
AlConnectionManager * al_ssl_connection_manager_new(int port, const char * keyfile, const char * dh_file, const char * ca_list, const char * password)
{
	AlConnectionManager * connection = al_new(AlConnectionManager);
	AlSslManager        * ssl_manager = al_new(AlSslManager);
 

	connection->wait = al_ssl_connection_wait;
	connection->close = al_ssl_connection_close;
	connection->read = al_ssl_connection_read;
	connection->send = al_ssl_connection_send;
	connection->del = al_ssl_connection_manager_delete;
	connection->conn = al_ssl_connection_connect;

	connection->data = ssl_manager;
	connection->fork = 1;

	ssl_manager->s_server_session_id_context = 1;
	ssl_manager->s_server_auth_session_id_context = 2;
	
	ssl_manager->port = port;
	ssl_manager->init = 0;
	ssl_manager->keyfile = al_strdup(keyfile);
	ssl_manager->dh_file = al_strdup(dh_file);
	ssl_manager->ca_list = al_strdup(ca_list);
	ssl_manager->password = al_strdup(password);


	return connection;
}


/**
 */
AlConnection * al_ssl_connection_wait(AlConnectionManager * connection_manager)
{
	AlConnection  * ret = NULL;
	AlSsl         * ret_ssl = NULL;
	AlSslManager  * ssl_manager;
	int             err;
	long            l_err;
	char            err_msg[200]; 


	al_return_val_if_fail(connection_manager, NULL);

	ssl_manager = connection_manager->data;

	if (!ssl_manager->init)
	{

		/* Build our SSL context*/
		ssl_manager->ctx = al_ssl_initialize_ctx(ssl_manager->keyfile, ssl_manager->ca_list, ssl_manager->password);
		al_ssl_load_dh_params(ssl_manager->ctx, ssl_manager->dh_file);

		SSL_CTX_set_session_id_context(ssl_manager->ctx,(void*)&(ssl_manager->s_server_session_id_context), sizeof(ssl_manager->s_server_session_id_context)); 
    
	/* Set our cipher list */
/*@fixme	if(ciphers)
	{
		SSL_CTX_set_cipher_list(ctx,ciphers);
	}
    */
        	SSL_CTX_set_verify(ssl_manager->ctx, SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT, 0);
	
	
		ssl_manager->sock = socket(PF_INET, SOCK_STREAM, 0);
		if (ssl_manager->sock == -1)
		{
			syslog(LOG_ERR, "Could not create the socket.\n");
			return NULL;
		}

		/* Give the socket a name. */
		ssl_manager->name.sin_family = AF_INET;
		ssl_manager->name.sin_port = htons (ssl_manager->port);
		ssl_manager->name.sin_addr.s_addr = htonl (INADDR_ANY);
		setsockopt(ssl_manager->sock, SOL_SOCKET, SO_REUSEADDR, &(ssl_manager->on), sizeof(ssl_manager->on));
		
		if (bind (ssl_manager->sock, (struct sockaddr *) &(ssl_manager->name), sizeof (ssl_manager->name)) < 0)
		{
			syslog (LOG_ERR, "Could not bind the server. Error %d\n", errno);
			return NULL;
		}

		if (listen (ssl_manager->sock, 5) < 0)
		{
			syslog (LOG_ERR, "Could not SSL start the listener.\n");
			return NULL;
		}

		syslog(LOG_NOTICE, "Acceptiong SSL connections on port %d\n", ssl_manager->port);

		ssl_manager->init = 1;
	}



	ret = al_new(AlConnection);
	ret->connection_manager = connection_manager;

	ret_ssl = al_new(AlSsl);
	ret->data = ret_ssl;

	ret_ssl->sock = accept(ssl_manager->sock, (struct sockaddr *)&(ret_ssl->sn), &(ret_ssl->snsize));

	if(ret_ssl->sock == -1)
	{
		al_delete(ret);
		al_delete(ret_ssl);

		syslog(LOG_NOTICE, "Close connections.\n");

		return NULL;
	}

	ret_ssl->ssl = SSL_new(ssl_manager->ctx);
	ret_ssl->sbio = BIO_new_socket(ret_ssl->sock, BIO_NOCLOSE);
        SSL_set_bio(ret_ssl->ssl, ret_ssl->sbio, ret_ssl->sbio);

	if((err = SSL_accept(ret_ssl->ssl)) <= 0)
	{
		l_err = ERR_get_error();
		ERR_error_string(l_err, err_msg);
		syslog(LOG_ERR, "SSL accept error %d - %d - %s\n",SSL_get_error(ret_ssl->ssl,err), l_err, err_msg);

		al_delete(ret);
		al_delete(ret_ssl);

		return NULL;
        }

	ret_ssl->hostname = gethostbyaddr(&(ret_ssl->sn.sin_addr), sizeof(ret_ssl->sn.sin_addr), AF_INET);
	if (ret_ssl->hostname) syslog(LOG_NOTICE, "Accepted new SLL connection from %s.\n", ret_ssl->hostname->h_name);


	return ret;
}


/**
 */
void al_ssl_connection_close(AlConnection * connection)
{
	al_return_if_fail(connection);

	al_delete(connection->data);
	al_delete(connection);
}


/**
 */
AlString * al_ssl_connection_read(AlConnection * connection)
{
	AlString    * ret = NULL;
	AlString    * s = NULL;
	AlSsl       * ssl_sock;
	const int     len = 3;
	int           i;
	char          buf[len];
	int           rec;


	al_return_val_if_fail(connection, NULL);

	ssl_sock = connection->data;
	s = al_string_new();
	
	while ((rec = SSL_read(ssl_sock->ssl, buf, sizeof(char) * len)) > 0)
	{

		if (rec == 3 && buf[0] == 'E' && buf[1] == 'O' && buf[2] == 'P')
		{
			// End of request packet
			ret = s;
			break;
		}
		else if (rec == 3 && buf[0] == 'E' && buf[1] == 'O' && buf[2] == 'F')
		{
			// End of trasmisson
			ret = NULL;
			if (ssl_sock->hostname) syslog(LOG_NOTICE, "Close connection from %s.\n", ssl_sock->hostname->h_name);
			shutdown(ssl_sock->sock, SHUT_RDWR);
			break;
		}


		for (i=0;i < rec;i++) al_string_append_char(s, buf[i]);

	}


	if (ret == NULL && s)
	{
		// Incomplete data. discard.
		al_string_delete(s);
	}


	return ret;
}


/**
 */
void al_ssl_connection_send(AlConnection * connection, const AlString * response)
{
	AlSsl * ssl_sock;


	al_return_if_fail(connection && response);

	ssl_sock = connection->data;

	if (SSL_write(ssl_sock->ssl, al_string_get(response), al_string_len(response)) == -1)
	{
		syslog(LOG_ERR, "Could not send data %s\n", al_string_get(response));
	}

	if (SSL_write(ssl_sock->ssl, "EOP", strlen("EOP")) == -1)
	{
		syslog(LOG_ERR, "Could not send data %s\n", "EOP");
	}

}


/**
 */
AlConnection * al_ssl_connection_connect(AlConnectionManager * connection_manager, const char * server)
{
	AlSslManager   * ssl_manager;
	AlConnection   * ret;
	AlSsl          * ret_ssl;
	struct hostent * pent;
	int             err;
	long            l_err;
	char            err_msg[200];



	al_return_val_if_fail(connection_manager && server, NULL);
	
	ssl_manager = connection_manager->data;

	if (!ssl_manager->init)
	{

		/* Build our SSL context*/
		ssl_manager->ctx = al_ssl_initialize_ctx(ssl_manager->keyfile, ssl_manager->ca_list, ssl_manager->password);

		SSL_CTX_set_session_id_context(ssl_manager->ctx,(void*)&(ssl_manager->s_server_session_id_context), sizeof(ssl_manager->s_server_session_id_context)); 
    
        	SSL_CTX_set_verify(ssl_manager->ctx, SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT, 0);
	
		ssl_manager->init = 1;
	}


	ret = al_new(AlConnection);
	ret->connection_manager = connection_manager;
	ret_ssl = al_new(AlSsl);
	ret->data = ret_ssl;

	ret_ssl->snsize = sizeof(ret_ssl->sn);

	ret_ssl->sn.sin_family = AF_INET;

	/* Figure out the internet address. */
	ret_ssl->sn.sin_addr.s_addr = inet_addr(server);
	
	if (ret_ssl->sn.sin_addr.s_addr == INADDR_NONE)
	{
		pent = gethostbyname(server);
		if (pent)
		{
			ret_ssl->sn.sin_addr.s_addr = *((unsigned long *) (pent->h_addr));
		}
		else
		{
			al_log (("Could not resolve the host %s\n", server));
			al_delete(ret_ssl);
			al_delete(ret);
			return NULL;
		}
	}

	ret_ssl->sock = socket(PF_INET, SOCK_STREAM, 0);
	if (ret_ssl->sock == -1)
	{
		al_log(("Could not create the socket.\n"));
		al_delete(ret_ssl);
		al_delete(ret);
		return NULL;
	}

	/* Figure out the port name. */
	ret_ssl->sn.sin_port =  htons(ssl_manager->port);

	if (connect(ret_ssl->sock, (struct sockaddr*)&ret_ssl->sn, ret_ssl->snsize) == -1)
	{
		al_log (("Could not connect to %s at port %d\n", server, ssl_manager->port));
		al_delete(ret_ssl);
		al_delete(ret);
		ret = NULL;
	}
	else
	{
		ret_ssl->ssl = SSL_new(ssl_manager->ctx);
		ret_ssl->sbio = BIO_new_socket(ret_ssl->sock, BIO_NOCLOSE);
        	SSL_set_bio(ret_ssl->ssl, ret_ssl->sbio, ret_ssl->sbio);

		if ((err = SSL_connect(ret_ssl->ssl)) <= 0)
		{
			l_err = ERR_get_error();
			ERR_error_string(l_err, err_msg);
			fprintf(stderr,"SSL connect error %d - %d - %s\n",SSL_get_error(ret_ssl->ssl,err), l_err, err_msg);
			al_delete(ret_ssl);
			al_delete(ret);
			ret = NULL;
		}
		else
		{
			al_ssl_check_cert(ret_ssl->ssl, server);
		}
		
	}


	return ret;
}


/**
 */
SSL_CTX * al_ssl_initialize_ctx(const char * keyfile, const char * ca_list, const char * password)
{
	SSL_METHOD * meth;
	SSL_CTX    * ctx;


	/* Global system initialization*/
	SSL_library_init();
	SSL_load_error_strings();

	/* Set up a SIGPIPE handler */
	signal(SIGPIPE,al_ssl_sigpipe_handle);

	/* Create our context*/
	meth = SSLv23_method();
	ctx = SSL_CTX_new(meth);

	/* Load our keys and certificates*/
	if(!(SSL_CTX_use_certificate_chain_file(ctx, keyfile)))
		fprintf(stderr, "Can't read certificate file %s.\n", keyfile);

	al_ssl_pass = al_strdup(password);
	SSL_CTX_set_default_passwd_cb(ctx, al_ssl_password_cb);
	if(!(SSL_CTX_use_PrivateKey_file(ctx, keyfile,SSL_FILETYPE_PEM)))
		fprintf(stderr, "Can't read key file %s.\n", keyfile);

	if (ca_list)
	{
		/* Load the CAs we trust*/
		if(!(SSL_CTX_load_verify_locations(ctx, ca_list,0)))
			fprintf(stderr, "Can't read CA list %s.\n", ca_list);
	}

#if (OPENSSL_VERSION_NUMBER < 0x00905100L)
	SSL_CTX_set_verify_depth(ctx,1);
#endif

	return ctx;
}
 

/**
 */
void al_ssl_destroy_ctx(SSL_CTX *ctx)
{
	SSL_CTX_free(ctx);
}


/**
 * The password code is not thread safe
 */
static int al_ssl_password_cb(char *buf, int num, int rwflag, void * userdata)
{
	if(num<strlen(al_ssl_pass)+1)
		return 0;

	strcpy(buf, al_ssl_pass);

	return strlen(al_ssl_pass);
}


/**
 */
static void al_ssl_sigpipe_handle(int x)
{
}


/**
 */
void al_ssl_check_cert(SSL * ssl, const char * host)
{
	X509 * peer;
	char   peer_CN[256];


	if(SSL_get_verify_result(ssl) != X509_V_OK)
		al_exit(1, "Certificate doesn't verify (host %s)\n", host);

	/**
	 * Check the cert chain. The chain length is automatically checked by OpenSSL when
	 * we set the verify depth in the ctx.
         */

	/* Check the common name*/
	peer = SSL_get_peer_certificate(ssl);
	X509_NAME_get_text_by_NID(X509_get_subject_name(peer), NID_commonName, peer_CN, 256);
	if(strcasecmp(peer_CN,host))
		al_exit(stderr, "Common name doesn't match host name (host %s)\n", host);

}


/**
 */
void al_ssl_load_dh_params(SSL_CTX *ctx, const char * file)
{
	DH  * ret=0;
	BIO * bio;

	if (file == NULL) return;

	if ((bio=BIO_new_file(file,"r")) == NULL)
		fprintf(stderr, "Couldn't open DH file %s\n", file);

	ret=PEM_read_bio_DHparams(bio, NULL, NULL, NULL);
	BIO_free(bio);

	if(SSL_CTX_set_tmp_dh(ctx,ret) < 0)
		fprintf(stderr, "Couldn't set DH parameters\n");

}

